//
// Copyright 2025 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_EXEC_VDF_INDEXED_WEIGHTS_H
#define PXR_EXEC_VDF_INDEXED_WEIGHTS_H

/// \file

#include "pxr/pxr.h"

#include "pxr/exec/vdf/indexedData.h"

PXR_NAMESPACE_OPEN_SCOPE

using VdfIndexedWeights = VdfIndexedData<float>;

PXR_NAMESPACE_CLOSE_SCOPE

#endif
