//
// Copyright 2016 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_BASE_GF_DECLARE_H
#define PXR_BASE_GF_DECLARE_H

#include "pxr/pxr.h"

/// \file gf/declare.h
/// Declares Gf types.
///
/// This file provides forward declarations for Gf types.

PXR_NAMESPACE_OPEN_SCOPE

class GfBBox3d;
class GfDualQuatd;
class GfDualQuatf;
class GfDualQuath;
class GfFrustum;
class GfInterval;
class GfMultiInterval;
class GfLine;
class GfLineSeg;
class GfPlane;
class GfQuatd;
class GfQuatf;
class GfQuath;
class GfQuaternion;
class GfRay;
class GfRect2i;
class GfRect2i;
class GfRotation;
class GfSize2;
class GfSize3;
class GfMatrix2d;
class GfMatrix2f;
class GfMatrix3d;
class GfMatrix3f;
class GfMatrix4d;
class GfMatrix4f;
class GfRange1d;
class GfRange1f;
class GfRange2d;
class GfRange2f;
class GfRange3d;
class GfRange3f;
class GfVec2d;
class GfVec2f;
class GfVec2h;
class GfVec2i;
class GfVec3d;
class GfVec3f;
class GfVec3h;
class GfVec3i;
class GfVec4d;
class GfVec4f;
class GfVec4h;
class GfVec4i;

PXR_NAMESPACE_CLOSE_SCOPE

#endif // PXR_BASE_GF_DECLARE_H
