#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
from pxr import Tf

Tf.ScriptModuleLoader()._LoadModulesForLibrary('LoadsUnknown')

from . import testTfScriptModuleLoader_Unknown
