#
# Copyright 2024 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
"""
Level of Detail (LOD) API schemas for USD.
"""

from pxr import Tf
Tf.PreparePythonModule()
del Tf
