//
// Copyright 2016 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_USD_IMAGING_USD_IMAGING_DEBUG_CODES_H
#define PXR_USD_IMAGING_USD_IMAGING_DEBUG_CODES_H

#include "pxr/pxr.h"
#include "pxr/base/tf/debug.h"

PXR_NAMESPACE_OPEN_SCOPE


TF_DEBUG_CODES(

    USDIMAGING_COLLECTIONS,
    USDIMAGING_COORDSYS,
    USDIMAGING_CHANGES,
    USDIMAGING_COMPUTATIONS,
    USDIMAGING_INSTANCER,
    USDIMAGING_MATERIAL_BINDING_RESOLUTION,
    USDIMAGING_PLUGINS,
    USDIMAGING_POINT_INSTANCER_PROTO_CREATED,
    USDIMAGING_POINT_INSTANCER_PROTO_CULLING,
    USDIMAGING_POPULATION,
    USDIMAGING_SELECTION,
    USDIMAGING_SHADERS,
    USDIMAGING_UPDATES

);


PXR_NAMESPACE_CLOSE_SCOPE

#endif // PXR_USD_IMAGING_USD_IMAGING_DEBUG_CODES_H
