//
// Copyright 2023 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_IMAGING_HD_VECTOR_SCHEMA_TYPE_DEFS_H
#define PXR_IMAGING_HD_VECTOR_SCHEMA_TYPE_DEFS_H

/// \deprecated
/// Use schemaTypeDefs.h header instead.

#include "pxr/imaging/hd/schemaTypeDefs.h"

#endif
