//
// Copyright 2025 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_EXEC_EXEC_DEBUG_CODES_H
#define PXR_EXEC_EXEC_DEBUG_CODES_H

#include "pxr/pxr.h"

#include "pxr/base/tf/debug.h"

PXR_NAMESPACE_OPEN_SCOPE

TF_DEBUG_CODES(
    EXEC_REQUEST_EXPIRATION,
    EXEC_REQUEST_INVALIDATION
);

PXR_NAMESPACE_CLOSE_SCOPE

#endif
