//
// Copyright 2019 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_USD_SDF_TIME_CODE_H
#define PXR_USD_SDF_TIME_CODE_H

/// \file sdf/timeCode.h

#include "pxr/pxr.h"
#include "pxr/base/gf/timeCode.h"

#include "pxr/base/vt/traits.h"

/// \deprecated SdfTimeCode is deprecated and has been replaced by
/// GfTimeCode. This preprocessor macro will be removed in a future release.
#define SdfTimeCode GfTimeCode

#endif // PXR_USD_SDF_TIME_CODE_H
