#
# Copyright 2018 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

"""Python bindings for libSdr"""

from pxr import Tf
Tf.PreparePythonModule()
del Tf
