//
// Copyright 2022 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_IMAGING_HGI_GL_DEBUG_CODES_H
#define PXR_IMAGING_HGI_GL_DEBUG_CODES_H

/// \file hgiGL/debugCodes.h

#include "pxr/pxr.h"
#include "pxr/base/tf/debug.h"

PXR_NAMESPACE_OPEN_SCOPE


TF_DEBUG_CODES(

    HGIGL_DEBUG_FRAMEBUFFER_CACHE

);


PXR_NAMESPACE_CLOSE_SCOPE

#endif
