//
// Copyright 2020 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_USD_AR_AR_H
#define PXR_USD_AR_AR_H

#define AR_VERSION 2

#endif
