//
// Copyright 2024 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//

// WARNING: THIS FILE IS GENERATED BY usdGenSchema.  DO NOT EDIT.
{% for pythonModule in pythonModules %}
TF_WRAP({{ pythonModule }});
{% endfor %}