#
# Copyright 2016 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#
"""
Graphics Foundation
This package defines classes for fundamental graphics types and operations.
"""

from pxr import Tf
Tf.PreparePythonModule()
del Tf
