#
# Copyright 2018 Pixar
#
# Licensed under the terms set forth in the LICENSE.txt file available at
# https://openusd.org/license.
#

# This file makes this module importable from python, which avoids a warning 
# when initializing the SdrRegistry.

