//
// Copyright 2024 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#include "pxr/pxr.h"
#include "pxr/base/tf/pyModule.h"

PXR_NAMESPACE_USING_DIRECTIVE

TF_WRAP_MODULE
{
    // Include non-schema classes here in the format TF_WRAP(CPP_class_name);
    
    // Generated Schema classes.  Do not remove or edit the following line.
    #include "generatedSchema.module.h"
}