//
// Copyright 2016 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_IMAGING_GARCH_GL_H
#define PXR_IMAGING_GARCH_GL_H

#include "pxr/pxr.h"
#include "pxr/base/arch/defines.h"

#include "pxr/imaging/garch/glApi.h"

#endif // PXR_IMAGING_GARCH_GL_H
