//
// Copyright 2021 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_IMAGING_HGI_DEBUG_CODES_H
#define PXR_IMAGING_HGI_DEBUG_CODES_H

/// \file hgi/debugCodes.h

#include "pxr/pxr.h"
#include "pxr/base/tf/debug.h"

PXR_NAMESPACE_OPEN_SCOPE


TF_DEBUG_CODES(

    HGI_DEBUG_DEVICE_CAPABILITIES,
    HGI_DEBUG_INSTANCE_CREATION,
    HGI_DEBUG_IS_SUPPORTED
);


PXR_NAMESPACE_CLOSE_SCOPE

#endif
