//
// Copyright 2024 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
#ifndef PXR_IMAGING_HDMTLX_DEBUG_CODES_H
#define PXR_IMAGING_HDMTLX_DEBUG_CODES_H

#include "pxr/pxr.h"
#include "pxr/base/tf/debug.h"

PXR_NAMESPACE_OPEN_SCOPE

TF_DEBUG_CODES(
    HDMTLX_DOCUMENT,
    HDMTLX_VERSION_UPGRADE,
    HDMTLX_WRITE_DOCUMENT,
    HDMTLX_WRITE_DOCUMENT_WITHOUT_INCLUDES
);

PXR_NAMESPACE_CLOSE_SCOPE

#endif // PXR_IMAGING_HDMTLX_DEBUG_CODES_H
